!MNH_LIC Copyright 1994-2022 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt
!MNH_LIC for details. version 1.
!-------------------------------------------------------------------------------
!#####################
MODULE MODI_ADV_BOUNDARIES
!#####################

use mode_mppdb

implicit none

#ifdef MNH_OPENACC
INTERFACE ADV_BOUNDARIES_DEVICE
  MODULE PROCEDURE ADV_BOUNDARIES_DEVICE1, ADV_BOUNDARIES_DEVICE2, ADV_BOUNDARIES_DEVICE3
END INTERFACE
#endif
!
!INTERFACE
!
#if 0
      SUBROUTINE ADV_BOUNDARIES_DEVICE1 ( HLBCX,HLBCY,PFIELD )
!
CHARACTER(LEN=4), DIMENSION(2), INTENT(IN)      :: HLBCX,HLBCY   ! X and Y-direc. LBC type
REAL, DIMENSION(:,:,:),   INTENT(INOUT)         :: PFIELD
!
END SUBROUTINE ADV_BOUNDARIES_DEVICE1
!
      SUBROUTINE ADV_BOUNDARIES_DEVICE2 ( HLBCX,HLBCY,PFIELD,PFIELDI )
!
CHARACTER(LEN=4), DIMENSION(2), INTENT(IN)      :: HLBCX,HLBCY   ! X and Y-direc. LBC type
REAL, DIMENSION(:,:,:),   INTENT(INOUT)         :: PFIELD
REAL, DIMENSION(:,:,:),   INTENT(IN)            :: PFIELDI
!
END SUBROUTINE ADV_BOUNDARIES_DEVICE2
!
      SUBROUTINE ADV_BOUNDARIES_DEVICE3 ( HLBCX,HLBCY,PFIELD,PFIELDI,HFIElD )
!
CHARACTER(LEN=4), DIMENSION(2), INTENT(IN)      :: HLBCX,HLBCY   ! X and Y-direc. LBC type
REAL, DIMENSION(:,:,:),   INTENT(INOUT)         :: PFIELD
REAL, DIMENSION(:,:,:),   INTENT(IN)            :: PFIELDI
CHARACTER(LEN=1),         INTENT(IN)            :: HFIELD  ! Field type
!
END SUBROUTINE ADV_BOUNDARIES_DEVICE3
!
      SUBROUTINE ADV_BOUNDARIES ( HLBCX,HLBCY,PFIELD,PFIELDI,HFIELD )
!
CHARACTER(LEN=4), DIMENSION(2), INTENT(IN)      :: HLBCX,HLBCY   ! X and Y-direc. LBC type
REAL, DIMENSION(:,:,:),   INTENT(INOUT)         :: PFIELD
REAL, DIMENSION(:,:,:),   INTENT(IN), OPTIONAL  :: PFIELDI
CHARACTER(LEN=1),         INTENT(IN), OPTIONAL  :: HFIELD  ! Field type
!
END SUBROUTINE ADV_BOUNDARIES
!
END INTERFACE
!
END MODULE MODI_ADV_BOUNDARIES
#endif
!
CONTAINS
!
#ifdef MNH_OPENACC
!     ####################################################################
      SUBROUTINE ADV_BOUNDARIES_DEVICE1 ( HLBCX,HLBCY,PFIELD )
!     ####################################################################
!
!!****  *ADV_BOUNDARIES* - routine to prepare the top and bottom Boundary Conditions
!!
!!
!!    AUTHOR
!!    ------
!!
!!
!-------------------------------------------------------------------------------
!
!*       0.    DECLARATIONS
!
USE MODD_PARAMETERS
USE MODE_ll
!
IMPLICIT NONE
!
!
!*       0.1   declarations of arguments
!
!
CHARACTER(LEN=4), DIMENSION(2), INTENT(IN)      :: HLBCX,HLBCY   ! X and Y-direc. LBC type
REAL, DIMENSION(:,:,:),   INTENT(INOUT)         :: PFIELD
!
!
!*       0.2   declarations of local variables
!
INTEGER             :: IKB       ! indice K Beginning in z direction
INTEGER             :: IKE       ! indice K End       in z direction
INTEGER             :: IIU, IJU  ! Index End in X and Y directions
!
!-------------------------------------------------------------------------------
IF (SIZE(PFIELD)==0) RETURN

!$acc data present( PFIELD )

CALL MPPDB_CHECK(PFIELD,"ADV_BOUNDARIES beg:PFIELD")
!
!*       1.    COMPUTE DIMENSIONS OF ARRAYS AND OTHER INDICES:
!              ----------------------------------------------
IKB = 1 + JPVEXT
IKE = SIZE(PFIELD,3) - JPVEXT
IIU=SIZE(PFIELD,1)
IJU=SIZE(PFIELD,2)
!
!
!
!-------------------------------------------------------------------------------
!
!*       2.    UPPER AND LOWER BC FILLING:
!              ---------------------------
!
!*       2.1    COMPUTE THE FIELD EXTRAPOLATIONS AT THE GROUND
!
!$acc kernels
   PFIELD  (:,:,IKB-1)   = PFIELD  (:,:,IKB)
!
!*       2.2    COMPUTE THE FIELD EXTRAPOLATIONS AT THE TOP
!
  PFIELD  (:,:,IKE+1)   = PFIELD  (:,:,IKE)
!$acc end kernels
!
!Not enough? !$acc update self(PFIELD(:,:,IKB-1))
!Not enough? !$acc update self(PFIELD(:,:,IKE+1))
! acc update self(PFIELD(:,:,:))
!
CALL MPPDB_CHECK(PFIELD,"ADV_BOUNDARIES end:PFIELD")
!
!$acc end data
!-------------------------------------------------------------------------------
!
END SUBROUTINE ADV_BOUNDARIES_DEVICE1
!
!     ####################################################################
      SUBROUTINE ADV_BOUNDARIES_DEVICE2 ( HLBCX,HLBCY,PFIELD,PFIELDI )
!     ####################################################################
!
!!****  *ADV_BOUNDARIES* - routine to prepare the top and bottom Boundary Conditions
!!
!!
!!    AUTHOR
!!    ------
!!
!!
!-------------------------------------------------------------------------------
!
!*       0.    DECLARATIONS
!
USE MODD_PARAMETERS
USE MODE_ll
!
IMPLICIT NONE
!
!
!*       0.1   declarations of arguments
!
!
CHARACTER(LEN=4), DIMENSION(2), INTENT(IN)      :: HLBCX,HLBCY   ! X and Y-direc. LBC type
REAL, DIMENSION(:,:,:),   INTENT(INOUT)         :: PFIELD
REAL, DIMENSION(:,:,:),   INTENT(IN)            :: PFIELDI
!
!
!*       0.2   declarations of local variables
!
INTEGER :: IKB       ! indice K Beginning in z direction
INTEGER :: IKE       ! indice K End       in z direction
INTEGER :: IIU, IJU  ! Index End in X and Y directions
INTEGER :: IIB,IIE,IJB,IJE ! interior domaine bound
LOGICAL :: GNORTH, GSOUTH, GWEST, GEAST
!
!-------------------------------------------------------------------------------
IF (SIZE(PFIELD)==0) RETURN

!$acc data present( PFIELD, PFIELDI )

CALL MPPDB_CHECK(PFIELDI,"ADV_BOUNDARIES beg:PFIELDI")
CALL MPPDB_CHECK(PFIELD,"ADV_BOUNDARIES beg:PFIELD")
!
!*       1.    COMPUTE DIMENSIONS OF ARRAYS AND OTHER INDICES:
!              ----------------------------------------------
CALL GET_INDICE_ll (IIB,IJB,IIE,IJE)
IKB = 1 + JPVEXT
IKE = SIZE(PFIELD,3) - JPVEXT
IIU=SIZE(PFIELD,1)
IJU=SIZE(PFIELD,2)

GWEST  = ( HLBCX(1) == 'OPEN' .AND. LWEST_ll()  )
GEAST  = ( HLBCX(2) == 'OPEN' .AND. LEAST_ll()  )
GSOUTH = ( HLBCY(1) == 'OPEN' .AND. LSOUTH_ll() )
GNORTH = ( HLBCY(2) == 'OPEN' .AND. LNORTH_ll() )
!
!
!$acc kernels
!
!-------------------------------------------------------------------------------
!
!*       2.    UPPER AND LOWER BC FILLING:
!              ---------------------------
!
!*       2.1    COMPUTE THE FIELD EXTRAPOLATIONS AT THE GROUND
!
!
   PFIELD  (:,:,IKB-1)   = PFIELD  (:,:,IKB)

!
!*       2.2    COMPUTE THE FIELD EXTRAPOLATIONS AT THE TOP
!
  PFIELD  (:,:,IKE+1)   = PFIELD  (:,:,IKE)
!
!
!*       3.    LATERAL BC FILLING
!              ---------------------------
!
  IF ( GWEST ) THEN
     PFIELD(:IIB-1,:,:) = PFIELDI(:IIB-1,:,:)
  END IF
  IF ( GEAST ) THEN
     PFIELD(IIE+1:,:,:) = PFIELDI(IIE+1:,:,:)
  END IF
  IF ( GSOUTH ) THEN
     PFIELD(:,:IJB-1,:) = PFIELDI(:,:IJB-1,:)
  END IF
  IF ( GNORTH ) THEN
     PFIELD(:,IJE+1:,:) = PFIELDI(:,IJE+1:,:)
  END IF
!$acc end kernels
!
#if 0
!Not enough?
  !$acc update self(PFIELD(:,:,IKB-1))
  !$acc update self(PFIELD(:,:,IKE+1))
  IF ( GWEST ) THEN
    !$acc update self(PFIELD(1,:,:))
  END IF
  IF ( GEAST ) THEN
    !$acc update self(PFIELD(IIU,:,:))
  END IF
  IF ( GSOUTH ) THEN
    !$acc update self(PFIELD(:,1,:))
  END IF
  IF ( GNORTH ) THEN
    !$acc update self(PFIELD(:,IJU,:))
  END IF
#else
! acc update self(PFIELD(:,:,:))
#endif
!
CALL MPPDB_CHECK(PFIELD,"ADV_BOUNDARIES end:PFIELD")
!
!$acc end data
!-------------------------------------------------------------------------------
!
END SUBROUTINE ADV_BOUNDARIES_DEVICE2
!
!     ####################################################################
      SUBROUTINE ADV_BOUNDARIES_DEVICE3 ( HLBCX,HLBCY,PFIELD,PFIELDI,HFIELD )
!     ####################################################################
!
!!****  *ADV_BOUNDARIES* - routine to prepare the top and bottom Boundary Conditions
!!
!!
!!    AUTHOR
!!    ------
!!
!!
!-------------------------------------------------------------------------------
!
!*       0.    DECLARATIONS
!
USE MODD_PARAMETERS
USE MODE_ll
!
IMPLICIT NONE
!
!
!*       0.1   declarations of arguments
!
!
CHARACTER(LEN=4), DIMENSION(2), INTENT(IN)      :: HLBCX,HLBCY   ! X and Y-direc. LBC type
REAL, DIMENSION(:,:,:),   INTENT(INOUT)         :: PFIELD
REAL, DIMENSION(:,:,:),   INTENT(IN)            :: PFIELDI
CHARACTER(LEN=1),         INTENT(IN)            :: HFIELD  ! Field type
!
!
!*       0.2   declarations of local variables
!
INTEGER :: IKB       ! indice K Beginning in z direction
INTEGER :: IKE       ! indice K End       in z direction
INTEGER :: IIU, IJU  ! Index End in X and Y directions
INTEGER :: IIB,IIE,IJB,IJE ! interior domaine bound
INTEGER :: IFLAG     ! Variable to workaround a performance problem with PGI compiler (at least 16.4)
LOGICAL :: GNORTH, GSOUTH, GWEST, GEAST
!
!-------------------------------------------------------------------------------
IF (SIZE(PFIELD)==0) RETURN

!$acc data present( PFIELD, PFIELDI )
!
CALL MPPDB_CHECK(PFIELDI,"ADV_BOUNDARIES beg:PFIELDI")
CALL MPPDB_CHECK(PFIELD,"ADV_BOUNDARIES beg:PFIELD")
!
!*       1.    COMPUTE DIMENSIONS OF ARRAYS AND OTHER INDICES:
!              ----------------------------------------------
CALL GET_INDICE_ll (IIB,IJB,IIE,IJE)
IKB = 1 + JPVEXT
IKE = SIZE(PFIELD,3) - JPVEXT
IIU=SIZE(PFIELD,1)
IJU=SIZE(PFIELD,2)

GWEST  = ( HLBCX(1) == 'OPEN' .AND. LWEST_ll()  )
GEAST  = ( HLBCX(2) == 'OPEN' .AND. LEAST_ll()  )
GSOUTH = ( HLBCY(1) == 'OPEN' .AND. LSOUTH_ll() )
GNORTH = ( HLBCY(2) == 'OPEN' .AND. LNORTH_ll() )

SELECT CASE (HFIELD)
  CASE ('U')
    IFLAG = 1
  CASE ('V')
    IFLAG = 2
  CASE ('W')
    IFLAG = 3
  CASE DEFAULT
    IFLAG = 0
END SELECT
!
!$acc kernels
!
!-------------------------------------------------------------------------------
!
!*       2.    UPPER AND LOWER BC FILLING:
!              ---------------------------
!
!*       2.1    COMPUTE THE FIELD EXTRAPOLATIONS AT THE GROUND
!
!
   IF (IFLAG==3) THEN !HFIELD=='W'
     PFIELD  (:,:,IKB  )   = PFIELDI (:,:,IKB)
   END IF
!
   PFIELD  (:,:,IKB-1)   = PFIELD  (:,:,IKB)

!
!*       2.2    COMPUTE THE FIELD EXTRAPOLATIONS AT THE TOP
!
  PFIELD  (:,:,IKE+1)   = PFIELD  (:,:,IKE)
!
!
!*       3.    LATERAL BC FILLING
!              ---------------------------
!
  IF ( GWEST ) THEN
     PFIELD(:IIB-1,:,:) = PFIELDI(:IIB-1,:,:)
     IF (IFLAG==1) & !HFIELD=='U'
     PFIELD(:IIB,:,:) = PFIELDI(:IIB,:,:)
  END IF
  IF ( GEAST ) THEN
     PFIELD(IIE+1:,:,:) = PFIELDI(IIE+1:,:,:)
  END IF
  IF ( GSOUTH ) THEN
     PFIELD(:,:IJB-1,:) = PFIELDI(:,:IJB-1,:)
     IF (IFLAG==2) & !HFIELD=='V'
     PFIELD(:,:IJB,:) = PFIELDI(:,:IJB,:)
  END IF
  IF ( GNORTH ) THEN
     PFIELD(:,IJE+1,:) = PFIELDI(:,IJE+1,:)
  END IF
!
!*       4. TOP BC for W
!
  IF (IFLAG==3) PFIELD(:,:,IKE+1) = 0. !IF HFIELD=='W'
!$acc end kernels
!
#if 0
!Not enough?
!add also if hfield =u or v
  !$acc update self(PFIELD(:,:,IKB-1))
  !$acc update self(PFIELD(:,:,IKE+1))
  IF ( GWEST ) THEN
    !$acc update self(PFIELD(1,:,:))
  END IF
  IF ( GEAST ) THEN
    !$acc update self(PFIELD(IIU,:,:))
  END IF
  IF ( GSOUTH ) THEN
    !$acc update self(PFIELD(:,1,:))
  END IF
  IF ( GNORTH ) THEN
    !$acc update self(PFIELD(:,IJU,:))
  END IF
#else
! acc update self(PFIELD(:,:,:))
#endif
!
CALL MPPDB_CHECK(PFIELD,"ADV_BOUNDARIES end:PFIELD")
!

!$acc end data
!-------------------------------------------------------------------------------
!
END SUBROUTINE ADV_BOUNDARIES_DEVICE3
#endif
!
!     ####################################################################
      SUBROUTINE ADV_BOUNDARIES ( HLBCX,HLBCY,PFIELD,PFIELDI,HFIELD )
!     ####################################################################
!
!!****  *ADV_BOUNDARIES* - routine to prepare the top and bottom Boundary Conditions 
!!
!!
!!    AUTHOR
!!    ------
!!   V.Masson
!! Correction :	
!!   J.Escobar : 15/09/2015 : WENO5 & JPHEXT <> 1
!!   C.Lac     : 10/16 : top BC for W 
!-------------------------------------------------------------------------------
!
!*       0.    DECLARATIONS
!         
USE MODD_PARAMETERS
USE MODE_ll
!
IMPLICIT NONE
!
!
!*       0.1   declarations of arguments
!
!
CHARACTER(LEN=4), DIMENSION(2), INTENT(IN)      :: HLBCX,HLBCY   ! X and Y-direc. LBC type
REAL, DIMENSION(:,:,:),   INTENT(INOUT)         :: PFIELD                        
REAL, DIMENSION(:,:,:),   INTENT(IN), OPTIONAL  :: PFIELDI                        
CHARACTER(LEN=1),         INTENT(IN), OPTIONAL  :: HFIELD  ! Field type
!
!
!*       0.2   declarations of local variables
!
INTEGER             :: IKB       ! indice K Beginning in z direction
INTEGER             :: IKE       ! indice K End       in z direction 
INTEGER             :: IIU, IJU  ! Index End in X and Y directions
INTEGER             :: IIB,IIE,IJB,IJE ! interior domaine bound
!
!-------------------------------------------------------------------------------
!
if ( present( PFIELDI ) ) call MPPDB_CHECK(PFIELDI,"ADV_BOUNDARIES beg:PFIELDI")
CALL MPPDB_CHECK(PFIELD,"ADV_BOUNDARIES beg:PFIELD")
!
!*       1.    COMPUTE DIMENSIONS OF ARRAYS AND OTHER INDICES:
!              ----------------------------------------------
CALL GET_INDICE_ll (IIB,IJB,IIE,IJE)
IKB = 1 + JPVEXT
IKE = SIZE(PFIELD,3) - JPVEXT
IIU=SIZE(PFIELD,1)
IJU=SIZE(PFIELD,2)
!
IF (SIZE(PFIELD)==0) RETURN
!-------------------------------------------------------------------------------
!
!*       2.    UPPER AND LOWER BC FILLING:   
!              ---------------------------
!
!*       2.1    COMPUTE THE FIELD EXTRAPOLATIONS AT THE GROUND
!
!
   IF (PRESENT(HFIELD) .AND. PRESENT(PFIELDI)) THEN
     IF (HFIELD=='W') &
     PFIELD  (:,:,IKB  )   = PFIELDI (:,:,IKB) 
   END IF
!
   PFIELD  (:,:,IKB-1)   = PFIELD  (:,:,IKB) 

!
!*       2.2    COMPUTE THE FIELD EXTRAPOLATIONS AT THE TOP
!
  PFIELD  (:,:,IKE+1)   = PFIELD  (:,:,IKE) 
!
!
!*       3.    LATERAL BC FILLING                                
!              ---------------------------
!
IF( PRESENT(PFIELDI) )  THEN
  IF (HLBCX(1)=='OPEN' .AND. LWEST_ll()) THEN
     PFIELD(:IIB-1,:,:) = PFIELDI(:IIB-1,:,:) ! 1
     IF (PRESENT(HFIELD)) THEN
       IF (HFIELD=='U') &
       PFIELD(:IIB,:,:) = PFIELDI(:IIB,:,:)   ! 2
     END IF
  END IF
  IF (HLBCX(2)=='OPEN' .AND. LEAST_ll()) THEN
     PFIELD(IIE+1:,:,:) = PFIELDI(IIE+1:,:,:) ! IIU
  END IF
  IF (HLBCY(1)=='OPEN' .AND. LSOUTH_ll()) THEN
     PFIELD(:,:IJB-1,:) = PFIELDI(:,:IJB-1,:) ! 1
     IF (PRESENT(HFIELD)) THEN
       IF (HFIELD=='V') &
       PFIELD(:,:IJB,:) = PFIELDI(:,:IJB,:) ! 2
     END IF
  END IF
  IF (HLBCY(2)=='OPEN' .AND. LNORTH_ll()) THEN
     PFIELD(:,IJE+1:,:) = PFIELDI(:,IJE+1:,:) ! IJU
  END IF
END IF
!
!*       4. TOP BC for W
!
 IF (PRESENT(HFIELD)) THEN
   IF (HFIELD=='W') PFIELD(:,:,IKE+1) = 0.
 END IF
!-------------------------------------------------------------------------------
CALL MPPDB_CHECK(PFIELD,"ADV_BOUNDARIES end:PFIELD")
!
END SUBROUTINE ADV_BOUNDARIES

END MODULE MODI_ADV_BOUNDARIES
