!MNH_LIC Copyright 2013-2022 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt
!MNH_LIC for details. version 1.
!-----------------------------------------------------------------
!     #####################
      MODULE MODI_ADVECTION_UVW_CEN
!     #####################
!
INTERFACE
      SUBROUTINE ADVECTION_UVW_CEN(HUVW_ADV_SCHEME,                &
                           HLBCX, HLBCY,                           &
                           PTSTEP, KTCOUNT,                        &
                           PUM, PVM, PWM,                          &
                           PDUM, PDVM, PDWM,                       &
                           PUT, PVT, PWT,                          &
                           PRHODJ, PDXX, PDYY, PDZZ, PDZX, PDZY,   &
                           PRUS,PRVS, PRWS,                        &
#ifndef MNH_OPENACC
                           TPHALO2MLIST                            )
#else
                           TPHALO2_UT,TPHALO2_VT,TPHALO2_WT )
#endif
!
USE MODD_ARGSLIST_ll, ONLY : HALO2LIST_ll
!
CHARACTER(LEN=6),         INTENT(IN)    :: HUVW_ADV_SCHEME
!
CHARACTER(LEN=4),DIMENSION(2),INTENT(IN):: HLBCX, HLBCY  ! X- and Y-direc LBC
!
REAL,                     INTENT(IN)    :: PTSTEP!  time step
INTEGER,                  INTENT(IN)    :: KTCOUNT
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PUM, PVM, PWM
                                                  ! Variables at t-dt
REAL, DIMENSION(:,:,:),   INTENT(INOUT) :: PDUM, PDVM, PDWM
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PUT , PVT  , PWT, PRHODJ
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PDXX,PDYY,PDZZ
REAL, DIMENSION(:,:,:),   INTENT(INOUT) :: PDZX,PDZY
                                                  !  metric coefficients
REAL, DIMENSION(:,:,:),   INTENT(INOUT) :: PRUS , PRVS  , PRWS
                                                  ! Sources terms 
!
! halo lists for 4th order advection
#ifndef MNH_OPENACC
TYPE(HALO2LIST_ll), POINTER :: TPHALO2MLIST ! momentum variables
#else
TYPE(HALO2LIST_ll), POINTER :: TPHALO2_UT,TPHALO2_VT,TPHALO2_WT
#endif
!
END SUBROUTINE ADVECTION_UVW_CEN
!
END INTERFACE
!
END MODULE MODI_ADVECTION_UVW_CEN
!     ##########################################################################
      SUBROUTINE ADVECTION_UVW_CEN(HUVW_ADV_SCHEME,                &
                           HLBCX, HLBCY,                           &
                           PTSTEP, KTCOUNT,                        &
                           PUM, PVM, PWM,                          &
                           PDUM, PDVM, PDWM,                       &
                           PUT, PVT, PWT,                          &
                           PRHODJ, PDXX, PDYY, PDZZ, PDZX, PDZY,   &
                           PRUS,PRVS, PRWS,                        &
#ifndef MNH_OPENACC
                           TPHALO2MLIST                            )
#else
                           TPHALO2_UT,TPHALO2_VT,TPHALO2_WT        )
#endif
!     ##########################################################################
!
!!****  *ADVECTION * - routine to call the specialized advection routines
!!
!!    PURPOSE
!!    -------
!!      The purpose of this routine is to control the advection routines.
!!    For that, it is first necessary to compute the metric coefficients
!!    and the contravariant components of the momentum.
!!
!!**  METHOD
!!    ------
!!      The advection of momenta is calculated using a centred (second order) 
!!    scheme. 
!!
!!    EXTERNAL
!!    --------
!!
!!    IMPLICIT ARGUMENTS
!!    ------------------
!!      NONE
!!
!!    REFERENCE
!!    ---------
!!      Book1 and book2 ( routine ADVECTION )
!!
!!    AUTHOR
!!    ------
!!	V. Masson        * Meteo France *
!!
!!    MODIFICATIONS
!!    -------------
!!      Original    01/2013  (from ADVECTION routine)
!!      Modif
!!      J.Escobar 21/03/2013: for HALOK comment all NHALO=1 test
!  P. Wautelet 20/05/2019: add name argument to ADDnFIELD_ll + new ADD4DFIELD_ll subroutine
!  P. Wautelet    02/2020: use the new data structures and subroutines for budgets
!-------------------------------------------------------------------------------
!
!*       0.    DECLARATIONS
!              ------------
!
USE MODD_ARGSLIST_ll, ONLY : LIST_ll, HALO2LIST_ll
USE MODD_CONF
use modd_budget,      only: lbudget_u, lbudget_v, lbudget_w, NBUDGET_U, NBUDGET_V, NBUDGET_W, tbudgets
USE MODD_GRID_n

use mode_budget,     only: Budget_store_init, Budget_store_end
USE MODE_ll
#ifdef MNH_OPENACC
USE MODE_DEVICE
USE MODE_MNH_ZWORK, ONLY: MNH_MEM_GET, MNH_MEM_POSITION_PIN, MNH_MEM_RELEASE
use mode_msg
#endif
use mode_mppdb

USE MODI_ADVECUVW_2ND
USE MODI_ADVECUVW_4TH
USE MODI_CONTRAV
#ifndef MNH_OPENACC
USE MODI_SHUMAN
#else
USE MODI_SHUMAN_DEVICE
#endif
!-------------------------------------------------------------------------------
!
IMPLICIT NONE
!
!*       0.1   Declarations of dummy arguments :
!
CHARACTER(LEN=6),         INTENT(IN)    :: HUVW_ADV_SCHEME
!
CHARACTER(LEN=4),DIMENSION(2),INTENT(IN):: HLBCX, HLBCY  ! X- and Y-direc LBC
!
REAL,                     INTENT(IN)    :: PTSTEP!  time step
INTEGER,                  INTENT(IN)    :: KTCOUNT
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PUM, PVM, PWM
                                                  ! Variables at t-dt
REAL, DIMENSION(:,:,:),   INTENT(INOUT) :: PDUM, PDVM, PDWM
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PUT , PVT  , PWT, PRHODJ
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PDXX,PDYY,PDZZ
REAL, DIMENSION(:,:,:),   INTENT(INOUT) :: PDZX,PDZY
                                                  !  metric coefficients
REAL, DIMENSION(:,:,:),   INTENT(INOUT) :: PRUS , PRVS  , PRWS
                                                  ! Sources terms 
!
! halo lists for 4th order advection
#ifndef MNH_OPENACC
TYPE(HALO2LIST_ll), POINTER :: TPHALO2MLIST ! momentum variables
#else
TYPE(HALO2LIST_ll), POINTER :: TPHALO2_UT,TPHALO2_VT,TPHALO2_WT
#endif
!
!
!*       0.2   declarations of local variables
!
!
INTEGER :: IIU, IJU, IKU
#ifndef MNH_OPENACC
REAL, DIMENSION(:,:,:), allocatable :: ZUS
REAL, DIMENSION(:,:,:), allocatable :: ZVS
REAL, DIMENSION(:,:,:), allocatable :: ZWS
                                                  ! guess of cartesian components of
                                                  ! momentum at future (+PTSTEP) timestep
REAL, DIMENSION(:,:,:), allocatable :: ZRUS
REAL, DIMENSION(:,:,:), allocatable :: ZRVS
REAL, DIMENSION(:,:,:), allocatable :: ZRWS
                                                  ! cartesian components of
                                                  ! rhodJ times the tendency of
                                                  ! momentum from previous (-PTSTEP)
                                                  ! to future (+PTSTEP) timestep
!
REAL, DIMENSION(:,:,:), allocatable :: ZRUT
REAL, DIMENSION(:,:,:), allocatable :: ZRVT
REAL, DIMENSION(:,:,:), allocatable :: ZRWT
                                                  ! cartesian
                                                  ! components of
                                                  ! momentum
!
REAL, DIMENSION(:,:,:), allocatable :: ZRUCT
REAL, DIMENSION(:,:,:), allocatable :: ZRVCT
REAL, DIMENSION(:,:,:), allocatable :: ZRWCT
                                                  ! contravariant
                                                  ! components
                                                  ! of momentum
REAL, DIMENSION(:,:,:), allocatable :: ZMXM_RHODJ
REAL, DIMENSION(:,:,:), allocatable :: ZMYM_RHODJ
REAL, DIMENSION(:,:,:), allocatable :: ZMZM_RHODJ
#else
REAL, DIMENSION(:,:,:), pointer, contiguous :: ZUS
REAL, DIMENSION(:,:,:), pointer, contiguous :: ZVS
REAL, DIMENSION(:,:,:), pointer, contiguous :: ZWS
                                                  ! guess of cartesian components of
                                                  ! momentum at future (+PTSTEP) timestep
REAL, DIMENSION(:,:,:), pointer, contiguous :: ZRUS
REAL, DIMENSION(:,:,:), pointer, contiguous :: ZRVS
REAL, DIMENSION(:,:,:), pointer, contiguous :: ZRWS
                                                  ! cartesian components of
                                                  ! rhodJ times the tendency of
                                                  ! momentum from previous (-PTSTEP)
                                                  ! to future (+PTSTEP) timestep
!
REAL, DIMENSION(:,:,:), pointer, contiguous :: ZRUT
REAL, DIMENSION(:,:,:), pointer, contiguous :: ZRVT
REAL, DIMENSION(:,:,:), pointer, contiguous :: ZRWT
                                                  ! cartesian
                                                  ! components of
                                                  ! momentum
!
REAL, DIMENSION(:,:,:), pointer, contiguous :: ZRUCT
REAL, DIMENSION(:,:,:), pointer, contiguous :: ZRVCT
REAL, DIMENSION(:,:,:), pointer, contiguous :: ZRWCT
                                                  ! contravariant
                                                  ! components
                                                  ! of momentum
REAL, DIMENSION(:,:,:), pointer, contiguous :: ZMXM_RHODJ
REAL, DIMENSION(:,:,:), pointer, contiguous :: ZMYM_RHODJ
REAL, DIMENSION(:,:,:), pointer, contiguous :: ZMZM_RHODJ
#endif
!
INTEGER                     :: IINFO_ll    ! return code of parallel routine
TYPE(LIST_ll), POINTER      :: TZFIELDS_ll ! list of fields to exchange
!
!-------------------------------------------------------------------------------
!$acc data present( PUM, PVM, PWM, PDUM, PDVM, PDWM, PUT, PVT, PWT, PRHODJ, PDXX, PDYY, PDZZ, PDZX, PDZY, PRUS, PRVS, PRWS )

IF (MPPDB_INITIALIZED) THEN
  !Check all IN arrays
  CALL MPPDB_CHECK(PUM,"ADVECTION_UVW_CEN beg:PUM")
  CALL MPPDB_CHECK(PVM,"ADVECTION_UVW_CEN beg:PVM")
  CALL MPPDB_CHECK(PWM,"ADVECTION_UVW_CEN beg:PWM")
  CALL MPPDB_CHECK(PDUM,"ADVECTION_UVW_CEN beg:PDUM")
  CALL MPPDB_CHECK(PDVM,"ADVECTION_UVW_CEN beg:PDVM")
  CALL MPPDB_CHECK(PDWM,"ADVECTION_UVW_CEN beg:PDWM")
  CALL MPPDB_CHECK(PUT,"ADVECTION_UVW_CEN beg:PUT")
  CALL MPPDB_CHECK(PVT,"ADVECTION_UVW_CEN beg:PVT")
  CALL MPPDB_CHECK(PWT,"ADVECTION_UVW_CEN beg:PWT")
  CALL MPPDB_CHECK(PRHODJ,"ADVECTION_UVW_CEN beg:PRHODJ")
  CALL MPPDB_CHECK(PDXX,"ADVECTION_UVW_CEN beg:PDXX")
  CALL MPPDB_CHECK(PDYY,"ADVECTION_UVW_CEN beg:PDYY")
  CALL MPPDB_CHECK(PDZZ,"ADVECTION_UVW_CEN beg:PDZZ")
  CALL MPPDB_CHECK(PDZX,"ADVECTION_UVW_CEN beg:PDZX")
  CALL MPPDB_CHECK(PDZY,"ADVECTION_UVW_CEN beg:PDZY")
  !Check all INOUT arrays
  CALL MPPDB_CHECK(PRUS,"ADVECTION_UVW_CEN beg:PRUS")
  CALL MPPDB_CHECK(PRVS,"ADVECTION_UVW_CEN beg:PRVS")
  CALL MPPDB_CHECK(PRWS,"ADVECTION_UVW_CEN beg:PRWS")
END IF

if ( lbudget_u ) call Budget_store_init( tbudgets(NBUDGET_U), 'ADV', prus(:, :, :) )
if ( lbudget_v ) call Budget_store_init( tbudgets(NBUDGET_V), 'ADV', prvs(:, :, :) )
if ( lbudget_w ) call Budget_store_init( tbudgets(NBUDGET_W), 'ADV', prws(:, :, :) )

IIU = SIZE( put, 1 )
IJU = SIZE( put, 2 )
IKU = SIZE( put, 3 )

#ifndef MNH_OPENACC
allocate( zus        ( iiu, iju, iku ) )
allocate( zvs        ( iiu, iju, iku ) )
allocate( zws        ( iiu, iju, iku ) )
allocate( zrus       ( iiu, iju, iku ) )
allocate( zrvs       ( iiu, iju, iku ) )
allocate( zrws       ( iiu, iju, iku ) )
allocate( zrut       ( iiu, iju, iku ) )
allocate( zrvt       ( iiu, iju, iku ) )
allocate( zrwt       ( iiu, iju, iku ) )
allocate( zruct      ( iiu, iju, iku ) )
allocate( zrvct      ( iiu, iju, iku ) )
allocate( zrwct      ( iiu, iju, iku ) )
allocate( zmxm_rhodj ( iiu, iju, iku ) )
allocate( zmym_rhodj ( iiu, iju, iku ) )
allocate( zmzm_rhodj ( iiu, iju, iku ) )
#else
!Pin positions in the pools of MNH memory
CALL MNH_MEM_POSITION_PIN( 'ADVECTION_UVW_CEN' )

CALL MNH_MEM_GET( zus,        iiu, iju, iku )
CALL MNH_MEM_GET( zvs,        iiu, iju, iku )
CALL MNH_MEM_GET( zws,        iiu, iju, iku )
CALL MNH_MEM_GET( zrus,       iiu, iju, iku )
CALL MNH_MEM_GET( zrvs,       iiu, iju, iku )
CALL MNH_MEM_GET( zrws,       iiu, iju, iku )
CALL MNH_MEM_GET( zrut,       iiu, iju, iku )
CALL MNH_MEM_GET( zrvt,       iiu, iju, iku )
CALL MNH_MEM_GET( zrwt,       iiu, iju, iku )
CALL MNH_MEM_GET( zruct,      iiu, iju, iku )
CALL MNH_MEM_GET( zrvct,      iiu, iju, iku )
CALL MNH_MEM_GET( zrwct,      iiu, iju, iku )
CALL MNH_MEM_GET( zmxm_rhodj, iiu, iju, iku )
CALL MNH_MEM_GET( zmym_rhodj, iiu, iju, iku )
CALL MNH_MEM_GET( zmzm_rhodj, iiu, iju, iku )

!$acc data present( zus, zvs, zws, zrus, zrvs, zrws, zrut, zrvt, zrwt, &
!$acc &             zruct, zrvct, zrwct, zmxm_rhodj, zmym_rhodj, zmzm_rhodj )
#endif

#ifdef MNH_OPENACC
CALL INIT_ON_HOST_AND_DEVICE(ZUS,-1e99,'ADVECTION_UVW_CEN::ZUS')
CALL INIT_ON_HOST_AND_DEVICE(ZVS,-2e99,'ADVECTION_UVW_CEN::ZVS')
CALL INIT_ON_HOST_AND_DEVICE(ZWS,-3e99,'ADVECTION_UVW_CEN::ZWS')
CALL INIT_ON_HOST_AND_DEVICE(ZRUS,-1e99,'ADVECTION_UVW_CEN::ZRUS')
CALL INIT_ON_HOST_AND_DEVICE(ZRVS,-2e99,'ADVECTION_UVW_CEN::ZRVS')
CALL INIT_ON_HOST_AND_DEVICE(ZRWS,-3e99,'ADVECTION_UVW_CEN::ZRWS')
CALL INIT_ON_HOST_AND_DEVICE(ZRUT,-1e99,'ADVECTION_UVW_CEN::ZRUT')
CALL INIT_ON_HOST_AND_DEVICE(ZRVT,-2e99,'ADVECTION_UVW_CEN::ZRVT')
CALL INIT_ON_HOST_AND_DEVICE(ZRWT,-3e99,'ADVECTION_UVW_CEN::ZRWT')
CALL INIT_ON_HOST_AND_DEVICE(ZRUCT,-1e98,'ADVECTION_UVW_CEN::ZRUCT')
CALL INIT_ON_HOST_AND_DEVICE(ZRVCT,-2e98,'ADVECTION_UVW_CEN::ZRVCT')
CALL INIT_ON_HOST_AND_DEVICE(ZRWCT,-3e98,'ADVECTION_UVW_CEN::ZRWCT')
CALL INIT_ON_HOST_AND_DEVICE(ZMXM_RHODJ,-1e97,'ADVECTION_UVW_CEN::ZMXM_RHODJ')
CALL INIT_ON_HOST_AND_DEVICE(ZMYM_RHODJ,-2e97,'ADVECTION_UVW_CEN::ZMYM_RHODJ')
CALL INIT_ON_HOST_AND_DEVICE(ZMZM_RHODJ,-3e97,'ADVECTION_UVW_CEN::ZMZM_RHODJ')
#endif

#ifndef MNH_OPENACC
ZMXM_RHODJ = MXM(PRHODJ)
ZMYM_RHODJ = MYM(PRHODJ)
ZMZM_RHODJ = MZM(PRHODJ)
#else
CALL MXM_DEVICE(PRHODJ,ZMXM_RHODJ)
CALL MYM_DEVICE(PRHODJ,ZMYM_RHODJ)
CALL MZM_DEVICE(PRHODJ,ZMZM_RHODJ)
#endif
!
!*       1.     COMPUTES THE CONTRAVARIANT COMPONENTS
!	        -------------------------------------
!
!$acc kernels present(ZRUT,ZRVT,ZRWT,PUT,PVT,PWT,ZMXM_RHODJ,ZMYM_RHODJ,ZMZM_RHODJ)
ZRUT(:,:,:) = PUT(:,:,:) * ZMXM_RHODJ(:,:,:)
ZRVT(:,:,:) = PVT(:,:,:) * ZMYM_RHODJ(:,:,:)
ZRWT(:,:,:) = PWT(:,:,:) * ZMZM_RHODJ(:,:,:)
!$acc end kernels
!
#ifndef MNH_OPENACC
IF (HUVW_ADV_SCHEME=='CEN2ND' ) THEN
  CALL CONTRAV (HLBCX,HLBCY,ZRUT,ZRVT,ZRWT,PDXX,PDYY,PDZZ,PDZX,PDZY,ZRUCT,ZRVCT,ZRWCT,2)
ELSEIF (HUVW_ADV_SCHEME=='CEN4TH') THEN
  CALL CONTRAV (HLBCX,HLBCY,ZRUT,ZRVT,ZRWT,PDXX,PDYY,PDZZ,PDZX,PDZY,ZRUCT,ZRVCT,ZRWCT,4)
END IF
#else
IF (HUVW_ADV_SCHEME=='CEN2ND' ) THEN
  CALL CONTRAV_DEVICE (HLBCX,HLBCY,ZRUT,ZRVT,ZRWT,PDXX,PDYY,PDZZ,PDZX,PDZY,ZRUCT,ZRVCT,ZRWCT,2, &
                 ODATA_ON_DEVICE=.TRUE.)
ELSEIF (HUVW_ADV_SCHEME=='CEN4TH') THEN
  CALL CONTRAV_DEVICE (HLBCX,HLBCY,ZRUT,ZRVT,ZRWT,PDXX,PDYY,PDZZ,PDZX,PDZY,ZRUCT,ZRVCT,ZRWCT,4, &
                 ODATA_ON_DEVICE=.TRUE.)
END IF
!Not necessary: already done in contrav_device !$acc update self(ZRUCT,ZRVCT,ZRWCT)
#endif
!
NULLIFY(TZFIELDS_ll)
!!$IF(NHALO == 1) THEN
  CALL ADD3DFIELD_ll( TZFIELDS_ll, ZRWCT, 'ADVECTION_UVW_CEN::ZRWCT' )
  CALL ADD3DFIELD_ll( TZFIELDS_ll, ZRUCT, 'ADVECTION_UVW_CEN::ZRUCT' )
  CALL ADD3DFIELD_ll( TZFIELDS_ll, ZRVCT, 'ADVECTION_UVW_CEN::ZRVCT' )
  CALL UPDATE_HALO_ll(TZFIELDS_ll,IINFO_ll)
  CALL CLEANLIST_ll(TZFIELDS_ll)
  !$acc update device(ZRUCT, ZRVCT, ZRWCT)
!!$END IF
!
!-------------------------------------------------------------------------------
!
!*       2.     TERM FROM PREVIOUS TIME-STEP (from initial_guess)
!	        ----------------------------
!
!$acc kernels present(ZRUS,ZRVS,ZRWS,PUM,PVM,PWM,ZMXM_RHODJ,ZMYM_RHODJ,ZMZM_RHODJ)
ZRUS(:,:,:)   = PUM(:,:,:)  * ZMXM_RHODJ(:,:,:)/(2.*PTSTEP)
ZRVS(:,:,:)   = PVM(:,:,:)  * ZMYM_RHODJ(:,:,:)/(2.*PTSTEP)
ZRWS(:,:,:)   = PWM(:,:,:)  * ZMZM_RHODJ(:,:,:)/(2.*PTSTEP)
!$acc end kernels
!
!-------------------------------------------------------------------------------
!
!*       3.     CALLS THE ADVECTION ROUTINES FOR THE MOMENTUM 
!	        ---------------------------------------------
!
! choose between 2nd and 4th order momentum advection.
IF (HUVW_ADV_SCHEME=='CEN2ND' ) THEN                                      
!
#ifdef MNH_OPENACC
  call Print_msg( NVERB_ERROR, 'GEN', 'ADVECTION_UVW_CEN', 'OpenACC: CEN2ND not yet implemented' )
#endif
   CALL ADVECUVW_2ND (PUT,PVT,PWT,ZRUCT,ZRVCT,ZRWCT,ZRUS,ZRVS,ZRWS)
!
ELSEIF (HUVW_ADV_SCHEME=='CEN4TH') THEN
! 
   CALL ADVECUVW_4TH ( HLBCX, HLBCY, ZRUCT, ZRVCT, ZRWCT,            &
                       PUT, PVT, PWT, ZRUS, ZRVS, ZRWS,              &
#ifndef MNH_OPENACC
                       TPHALO2MLIST                                  )
#else
                       TPHALO2_UT,TPHALO2_VT,TPHALO2_WT              )
#endif
!
END IF
!
!$acc kernels present(ZRUS,ZRVS,ZRWS,ZMXM_RHODJ,ZMYM_RHODJ,ZMZM_RHODJ) &
!$acc & present_cr(PRUS,PRVS,PRWS,PDUM,PDWM)
ZUS(:,:,:) = ZRUS(:,:,:)/ZMXM_RHODJ(:,:,:)*2.*PTSTEP
ZVS(:,:,:) = ZRVS(:,:,:)/ZMYM_RHODJ(:,:,:)*2.*PTSTEP
ZWS(:,:,:) = ZRWS(:,:,:)/ZMZM_RHODJ(:,:,:)*2.*PTSTEP
!-------------------------------------------------------------------------------
!
!*       5.     Extracts the variation between current and future time step
!	        -----------------------------------------------------------
!
PRUS(:,:,:) = PRUS(:,:,:) + ( ZUS(:,:,:) - PUM(:,:,:) - 0.5* PDUM(:,:,:)) * ZMXM_RHODJ(:,:,:)/(PTSTEP)
PRVS(:,:,:) = PRVS(:,:,:) + ( ZVS(:,:,:) - PVM(:,:,:) - 0.5* PDVM(:,:,:)) * ZMYM_RHODJ(:,:,:)/(PTSTEP)
PRWS(:,:,:) = PRWS(:,:,:) + ( ZWS(:,:,:) - PWM(:,:,:) - 0.5* PDWM(:,:,:)) * ZMZM_RHODJ(:,:,:)/(PTSTEP)
!
PDUM(:,:,:) = ZUS(:,:,:) - PUM(:,:,:)
PDVM(:,:,:) = ZVS(:,:,:) - PVM(:,:,:)
PDWM(:,:,:) = ZWS(:,:,:) - PWM(:,:,:)
!$acc end kernels

if ( lbudget_u ) call Budget_store_end( tbudgets(NBUDGET_U), 'ADV', prus(:, :, :) )
if ( lbudget_v ) call Budget_store_end( tbudgets(NBUDGET_V), 'ADV', prvs(:, :, :) )
if ( lbudget_w ) call Budget_store_end( tbudgets(NBUDGET_W), 'ADV', prws(:, :, :) )

IF (MPPDB_INITIALIZED) THEN
  !Check all INOUT arrays
  CALL MPPDB_CHECK(PRUS,"ADVECTION_UVW_CEN end:PRUS")
  CALL MPPDB_CHECK(PRVS,"ADVECTION_UVW_CEN end:PRVS")
  CALL MPPDB_CHECK(PRWS,"ADVECTION_UVW_CEN end:PRWS")
END IF

!$acc end data

#ifdef MNH_OPENACC
!Release all memory allocated with MNH_MEM_GET calls since last call to MNH_MEM_POSITION_PIN
CALL MNH_MEM_RELEASE( 'ADVECTION_UVW_CEN' )
#endif

!$acc end data

!-------------------------------------------------------------------------------
!
END SUBROUTINE ADVECTION_UVW_CEN
