!MNH_LIC Copyright 2013-2022 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt
!MNH_LIC for details. version 1.
!-----------------------------------------------------------------
!     ###########################
      MODULE MODI_ADVECUVW_WENO_K
!     ###########################
!
INTERFACE
!
      SUBROUTINE ADVECUVW_WENO_K(HLBCX, HLBCY, KWENO_ORDER, PUT, PVT, PWT,       &
                             PRUCT, PRVCT, PRWCT, PRUS, PRVS, PRWS, & 
#ifndef MNH_OPENACC
                             TPHALO2LIST )
#else
                             TPHALO2_UT,TPHALO2_VT,TPHALO2_WT )
#endif
!
USE MODD_ARGSLIST_ll, ONLY : HALO2LIST_ll
!
CHARACTER (LEN=4), DIMENSION(2), INTENT(IN) :: HLBCX ! X direction LBC type
CHARACTER (LEN=4), DIMENSION(2), INTENT(IN) :: HLBCY ! Y direction LBC type
INTEGER,                  INTENT(IN)    :: KWENO_ORDER   ! Order of the WENO
                                                         ! scheme (3 or 5)
!
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PRUCT ! contravariant
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PRVCT !  components
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PRWCT ! of momentum
!
REAL, DIMENSION(:,:,:),   INTENT(IN) :: PUT, PVT, PWT        ! U,V,W at t
!
REAL, DIMENSION(:,:,:),   INTENT(INOUT) :: PRUS, PRVS, PRWS     ! Source terms
!
#ifndef MNH_OPENACC
TYPE(HALO2LIST_ll), POINTER :: TPHALO2LIST ! list for diffusion
#else
TYPE(HALO2LIST_ll), POINTER :: TPHALO2_UT,TPHALO2_VT,TPHALO2_WT
#endif
!
!
END SUBROUTINE ADVECUVW_WENO_K
!
END INTERFACE
!
END MODULE MODI_ADVECUVW_WENO_K
!
!     ##########################################################################
      SUBROUTINE ADVECUVW_WENO_K(HLBCX, HLBCY, KWENO_ORDER, PUT, PVT, PWT,      &
                             PRUCT, PRVCT, PRWCT, PRUS, PRVS, PRWS, & 
#ifndef MNH_OPENACC
                             TPHALO2LIST )
#else
                             TPHALO2_UT,TPHALO2_VT,TPHALO2_WT )
#endif
!     ##########################################################################
!
!!    AUTHOR
!!    ------
!!
!!
!!    MODIFICATIONS
!!    ------------- 
!!      J.Escobar 21/03/2013: for HALOK comment all NHALO=1 tests
!!		  T.Lunet	 	02/10/2014: add get_halo for WENO 5
!!					suppress comment of NHALO=1 tests
!!      J.Escobar : 15/09/2015 : WENO5 & JPHEXT <> 1 
!!
!-------------------------------------------------------------------------------
!
!*       0.    DECLARATIONS
!              ------------
!
USE MODD_ARGSLIST_ll, ONLY : HALO2LIST_ll
USE MODD_CONF,        ONLY : L2D, NHALO
USE MODD_PARAMETERS

USE MODE_ll
#ifdef MNH_OPENACC
USE MODE_DEVICE
USE MODE_MNH_ZWORK,   ONLY: MNH_MEM_GET, MNH_MEM_POSITION_PIN, MNH_MEM_RELEASE
#endif
USE MODE_MPPDB

USE MODI_ADVEC_WENO_K_1_AUX
USE MODI_ADVEC_WENO_K_2_AUX
USE MODI_ADVEC_WENO_K_3_AUX
USE MODI_GET_HALO
#ifndef MNH_OPENACC
USE MODI_SHUMAN
#else
USE MODI_SHUMAN_DEVICE
#endif

IMPLICIT NONE
!
!*       0.1   Declarations of dummy arguments :
!
CHARACTER (LEN=4), DIMENSION(2), INTENT(IN) :: HLBCX ! X direction LBC type
CHARACTER (LEN=4), DIMENSION(2), INTENT(IN) :: HLBCY ! Y direction LBC type
INTEGER,                  INTENT(IN)    :: KWENO_ORDER   ! Order of the WENO
                                                         ! scheme (3 or 5)
!
REAL, DIMENSION(:,:,:),   INTENT(IN) :: PRUCT  ! contravariant
REAL, DIMENSION(:,:,:),   INTENT(IN) :: PRVCT  !  components
REAL, DIMENSION(:,:,:),   INTENT(IN) :: PRWCT  ! of momentum
!
REAL, DIMENSION(:,:,:),   INTENT(IN) :: PUT, PVT, PWT     ! Variables at t
!
REAL, DIMENSION(:,:,:),   INTENT(INOUT) :: PRUS, PRVS, PRWS     ! Source terms
!
#ifndef MNH_OPENACC
TYPE(HALO2LIST_ll), POINTER :: TPHALO2LIST ! list for diffusion
#else
TYPE(HALO2LIST_ll), POINTER :: TPHALO2_UT,TPHALO2_VT,TPHALO2_WT
#endif
!
!*       0.2   Declarations of local variables :
!
INTEGER                     :: IIU, IJU, IKU
TYPE(HALO2LIST_ll), POINTER :: TZHALO2_UT,TZHALO2_VT,TZHALO2_WT
INTEGER                     :: IINFO_ll    ! return code of parallel routine
!
#ifndef MNH_OPENACC
REAL, DIMENSION(:,:,:), allocatable :: ZMEAN, ZWORK
#else
REAL, DIMENSION(:,:,:), POINTER ,CONTIGUOUS :: ZMEAN, ZWORK
!
REAL, DIMENSION(:,:,:), POINTER ,CONTIGUOUS :: ZFPOS1, ZFPOS2, ZFPOS3
REAL, DIMENSION(:,:,:), POINTER ,CONTIGUOUS :: ZFNEG1, ZFNEG2, ZFNEG3
REAL, DIMENSION(:,:,:), POINTER ,CONTIGUOUS :: ZBPOS1, ZBPOS2, ZBPOS3
REAL, DIMENSION(:,:,:), POINTER ,CONTIGUOUS :: ZBNEG1, ZBNEG2, ZBNEG3
REAL, DIMENSION(:,:,:), POINTER ,CONTIGUOUS :: ZOMP1,  ZOMP2,  ZOMP3
REAL, DIMENSION(:,:,:), POINTER ,CONTIGUOUS :: ZOMN1,  ZOMN2,  ZOMN3
#endif
!
!
IF (MPPDB_INITIALIZED) THEN
  !Check all IN arrays
  CALL MPPDB_CHECK(PRUCT,"ADVECUVW_WENO_K beg:PRUCT")
  CALL MPPDB_CHECK(PRVCT,"ADVECUVW_WENO_K beg:PRVCT")
  CALL MPPDB_CHECK(PRWCT,"ADVECUVW_WENO_K beg:PRWCT")
  CALL MPPDB_CHECK(PUT,"ADVECUVW_WENO_K beg:PUT")
  CALL MPPDB_CHECK(PVT,"ADVECUVW_WENO_K beg:PVT")
  CALL MPPDB_CHECK(PWT,"ADVECUVW_WENO_K beg:PWT")
  !Check all INOUT arrays
  CALL MPPDB_CHECK(PRUS,"ADVECUVW_WENO_K beg:PRUS")
  CALL MPPDB_CHECK(PRVS,"ADVECUVW_WENO_K beg:PRVS")
  CALL MPPDB_CHECK(PRWS,"ADVECUVW_WENO_K beg:PRWS")
END IF

IIU = SIZE( PUT, 1 )
IJU = SIZE( PUT, 2 )
IKU = SIZE( PUT, 3 )

#ifndef MNH_OPENACC
allocate( ZMEAN(IIU, IJU, IKU) )
allocate( ZWORK(IIU, IJU, IKU) )
#else
!Pin positions in the pools of MNH memory
CALL MNH_MEM_POSITION_PIN()

CALL MNH_MEM_GET( ZMEAN, IIU, IJU, IKU )
CALL MNH_MEM_GET( ZWORK, IIU, IJU, IKU )
#endif

#ifdef MNH_OPENACC
CALL INIT_ON_HOST_AND_DEVICE(ZMEAN,1e90,'ADVECUVW_WENO_K::ZMEAN')
CALL INIT_ON_HOST_AND_DEVICE(ZWORK,2e90,'ADVECUVW_WENO_K::ZWORK')
#endif
!
!------------------------- ADVECTION OF MOMENTUM ------------------------------
!
!
#ifndef MNH_OPENACC
TZHALO2_UT => TPHALO2LIST                   ! 1rst add3dfield in model_n
TZHALO2_VT => TPHALO2LIST%NEXT              ! 2nd  add3dfield in model_n
TZHALO2_WT => TPHALO2LIST%NEXT%NEXT         ! 3rst add3dfield in model_n
#else
TZHALO2_UT => TPHALO2_UT
TZHALO2_VT => TPHALO2_VT
TZHALO2_WT => TPHALO2_WT
#endif
!
!      -------------------------------------------------------

!$acc data present( PRUCT, PRVCT, PRWCT, PUT, PVT, PWT, PRUS, PRVS, PRWS, ZMEAN, ZWORK )

SELECT CASE(KWENO_ORDER)
!
CASE(1) ! WENO 1
#ifndef MNH_OPENACC
!
!  U component
!
  PRUS = PRUS - DXM(UP_UX(PUT,MXF(PRUCT)))
!
  PRUS = PRUS - DYF(UP_MY(PUT,MXM(PRVCT)))
!
  PRUS = PRUS - DZF(UP_MZ(PUT,MXM(PRWCT)))
!
! V component
!
  PRVS = PRVS - DXF(UP_MX(PVT,MYM(PRUCT)))
!
  PRVS = PRVS - DYM(UP_VY(PVT,MYF(PRVCT)))
!
  PRVS = PRVS - DZF(UP_MZ(PVT,MYM(PRWCT)))
!
! W component
!
  PRWS = PRWS - DXF(UP_MX(PWT,MZM(PRUCT)))
!
  PRWS = PRWS - DYF(UP_MY(PWT,MZM(PRVCT)))
!
  PRWS = PRWS - DZM(UP_WZ(PWT,MZF(PRWCT)))
#else
!
!  U component
!
  !PRUS = PRUS - DXM(UP_UX(PUT,MXF(PRUCT)))
  CALL MXF_DEVICE(PRUCT,ZWORK)
  CALL UP_UX_DEVICE(PUT,ZWORK,ZMEAN)
  CALL DXM_DEVICE(ZMEAN,ZWORK)
!$acc kernels
  PRUS(:,:,:) = PRUS(:,:,:) - ZWORK(:,:,:)
!$acc end kernels
!
  !PRUS = PRUS - DYF(UP_MY(PUT,MXM(PRVCT)))
  CALL MXM_DEVICE(PRVCT,ZWORK)
  CALL UP_MY_DEVICE(PUT,ZWORK,ZMEAN)
  CALL DYF_DEVICE(ZMEAN,ZWORK)
!$acc kernels
  PRUS(:,:,:) = PRUS(:,:,:) - ZWORK(:,:,:)
!$acc end kernels
!
  !PRUS = PRUS - DZF(UP_MZ(PUT,MXM(PRWCT)))
  CALL MXM_DEVICE(PRWCT,ZWORK)
  CALL UP_MZ_DEVICE(PUT,ZWORK,ZMEAN)
  CALL DZF_DEVICE( ZMEAN, ZWORK )
!$acc kernels
  PRUS(:,:,:) = PRUS(:,:,:) - ZWORK(:,:,:)
!$acc end kernels
!
! V component
!
  !PRVS = PRVS - DXF(UP_MX(PVT,MYM(PRUCT)))
  CALL MYM_DEVICE(PRUCT,ZWORK)
  CALL UP_MX_DEVICE(PVT,ZWORK,ZMEAN)
  CALL DXF_DEVICE(ZMEAN,ZWORK)
!$acc kernels
  PRVS(:,:,:) = PRVS(:,:,:) - ZWORK(:,:,:)
!$acc end kernels
!
  !PRVS = PRVS - DYM(UP_VY(PVT,MYF(PRVCT)))
  CALL MYF_DEVICE(PRVCT,ZWORK)
  CALL UP_VY_DEVICE(PVT,ZWORK,ZMEAN)
  CALL DYM_DEVICE(ZMEAN,ZWORK)
!$acc kernels
  PRVS(:,:,:) = PRVS(:,:,:) - ZWORK(:,:,:)
!$acc end kernels
!
  !PRVS = PRVS - DZF( UP_MZ(PVT,MYM(PRWCT)))
  CALL MYM_DEVICE(PRWCT,ZWORK)
  CALL UP_MZ_DEVICE(PVT,ZWORK,ZMEAN)
  CALL DZF_DEVICE( ZMEAN, ZWORK )
!$acc kernels
  PRVS(:,:,:) = PRVS(:,:,:) - ZWORK(:,:,:)
!$acc end kernels
!
! W component
!
  !PRWS = PRWS - DXF(UP_MX(PWT,MZM(PRUCT)))
  CALL MZM_DEVICE(PRUCT,ZWORK)
  CALL UP_MX_DEVICE(PWT,ZWORK,ZMEAN)
  CALL DXF_DEVICE(ZMEAN,ZWORK)
!$acc kernels
  PRWS(:,:,:) = PRWS(:,:,:) - ZWORK(:,:,:)
!$acc end kernels
!
  !PRWS = PRWS - DYF(UP_MY(PWT,MZM(PRVCT)))
  CALL MZM_DEVICE(PRVCT,ZWORK)
  CALL UP_MY_DEVICE(PWT,ZWORK,ZMEAN)
  CALL DYF_DEVICE(ZMEAN,ZWORK)
!$acc kernels
  PRWS(:,:,:) = PRWS(:,:,:) - ZWORK(:,:,:)
!$acc end kernels
!
  !PRWS = PRWS - DZM(UP_WZ(PWT,MZF(PRWCT)))
  CALL MZF_DEVICE( PRWCT, ZWORK )
  CALL UP_WZ_DEVICE(PWT,ZWORK,ZMEAN)
  CALL DZM_DEVICE( ZMEAN, ZWORK )
!$acc kernels
  PRWS(:,:,:) = PRWS(:,:,:) - ZWORK(:,:,:)
!$acc end kernels
#endif
!
!
CASE(3) ! WENO 3
#ifndef MNH_OPENACC
!
! U component
!
  ZWORK = MXF(PRUCT)
  CALL ADVEC_WENO_K_2_UX(HLBCX, PUT, ZWORK, ZMEAN, TZHALO2_UT%HALO2)
  PRUS = PRUS - DXM(ZMEAN)
  
!   
  IF (.NOT.L2D) THEN
    ZWORK = MXM(PRVCT)
    CALL ADVEC_WENO_K_2_MY(HLBCY, PUT, ZWORK, ZMEAN, TZHALO2_UT%HALO2)
    PRUS = PRUS - DYF(ZMEAN)
  END IF
!
  PRUS = PRUS - DZF(WENO_K_2_MZ(PUT, MXM(PRWCT)))
!
! V component
!
  IF (.NOT.L2D) THEN
    ZWORK = MYM(PRUCT)
    CALL ADVEC_WENO_K_2_MX(HLBCX, PVT, ZWORK, ZMEAN, TZHALO2_VT%HALO2)
    PRVS = PRVS - DXF(ZMEAN)
!   
    ZWORK = MYF(PRVCT)
    CALL ADVEC_WENO_K_2_VY(HLBCY, PVT, ZWORK, ZMEAN, TZHALO2_VT%HALO2)
    PRVS = PRVS - DYM(ZMEAN)
!
    PRVS = PRVS - DZF(WENO_K_2_MZ(PVT, MYM(PRWCT)))
  END IF
!
! W component
!
  ZWORK = MZM(PRUCT)
  CALL ADVEC_WENO_K_2_MX(HLBCX, PWT, ZWORK, ZMEAN, TZHALO2_WT%HALO2)
  PRWS = PRWS - DXF(ZMEAN)
!
  IF (.NOT.L2D) THEN
    ZWORK = MZM(PRVCT)
    CALL ADVEC_WENO_K_2_MY(HLBCY, PWT, ZWORK, ZMEAN, TZHALO2_WT%HALO2)
    PRWS = PRWS - DYF(ZMEAN)
  END IF
!
  PRWS = PRWS - DZM(WENO_K_2_WZ(PWT,MZF(PRWCT)))
#else
  !Pin positions in the pools of MNH memory
  CALL MNH_MEM_POSITION_PIN()

  CALL MNH_MEM_GET( ZFPOS1, IIU, IJU, IKU )
  CALL MNH_MEM_GET( ZFPOS2, IIU, IJU, IKU )
  CALL MNH_MEM_GET( ZFNEG1, IIU, IJU, IKU )
  CALL MNH_MEM_GET( ZFNEG2, IIU, IJU, IKU )
  CALL MNH_MEM_GET( ZBPOS1, IIU, IJU, IKU )
  CALL MNH_MEM_GET( ZBPOS2, IIU, IJU, IKU )
  CALL MNH_MEM_GET( ZBNEG1, IIU, IJU, IKU )
  CALL MNH_MEM_GET( ZBNEG2, IIU, IJU, IKU )
  CALL MNH_MEM_GET( ZOMP1,  IIU, IJU, IKU )
  CALL MNH_MEM_GET( ZOMP2,  IIU, IJU, IKU )
  CALL MNH_MEM_GET( ZOMN1,  IIU, IJU, IKU )
  CALL MNH_MEM_GET( ZOMN2,  IIU, IJU, IKU )
!
! U component
!
  CALL MXF_DEVICE(PRUCT,ZWORK)
  CALL ADVEC_WENO_K_2_UX(HLBCX, PUT, ZWORK, ZMEAN, TZHALO2_UT%HALO2%WEST, TZHALO2_UT%HALO2%EAST, &
                         ZFPOS1(:,:,:), ZFPOS2(:,:,:), ZFNEG1(:,:,:), ZFNEG2(:,:,:),             &
                         ZBPOS1(:,:,:), ZBPOS2(:,:,:), ZBNEG1(:,:,:), ZBNEG2(:,:,:),             &
                         ZOMP1(:,:,:),  ZOMP2(:,:,:),  ZOMN1(:,:,:),  ZOMN2(:,:,:)               )
  CALL DXM_DEVICE(ZMEAN,ZWORK)
!$acc kernels
  PRUS(:,:,:) = PRUS(:,:,:) - ZWORK(:,:,:)
!$acc end kernels
!   
  IF (.NOT.L2D) THEN
    CALL MXM_DEVICE(PRVCT,ZWORK)
    CALL ADVEC_WENO_K_2_MY(HLBCY, PUT, ZWORK, ZMEAN, TZHALO2_UT%HALO2%NORTH, TZHALO2_UT%HALO2%SOUTH, &
                           ZFPOS1(:,:,:), ZFPOS2(:,:,:), ZFNEG1(:,:,:), ZFNEG2(:,:,:),               &
                           ZBPOS1(:,:,:), ZBPOS2(:,:,:), ZBNEG1(:,:,:), ZBNEG2(:,:,:),               &
                           ZOMP1(:,:,:),  ZOMP2(:,:,:),  ZOMN1(:,:,:),  ZOMN2(:,:,:)                 )
    CALL DYF_DEVICE(ZMEAN,ZWORK)
!$acc kernels
    PRUS(:,:,:) = PRUS(:,:,:) - ZWORK(:,:,:)
!$acc end kernels
  END IF
!
!  PRUS = PRUS - DZF(WENO_K_2_MZ(PUT, MXM(PRWCT)))
  CALL MXM_DEVICE(PRWCT,ZWORK)
  CALL WENO_K_2_MZ(PUT, ZWORK, ZMEAN,                                          &
                   ZFPOS1(:,:,:), ZFPOS2(:,:,:), ZFNEG1(:,:,:), ZFNEG2(:,:,:), &
                   ZBPOS1(:,:,:), ZBPOS2(:,:,:), ZBNEG1(:,:,:), ZBNEG2(:,:,:), &
                   ZOMP1(:,:,:),  ZOMP2(:,:,:),  ZOMN1(:,:,:),  ZOMN2(:,:,:)   )
  CALL DZF_DEVICE( ZMEAN, ZWORK )
!$acc kernels
  PRUS(:,:,:) = PRUS(:,:,:) - ZWORK(:,:,:)
!$acc end kernels
!
! V component
!
  IF (.NOT.L2D) THEN
    CALL MYM_DEVICE(PRUCT,ZWORK)
    CALL ADVEC_WENO_K_2_MX(HLBCX, PVT, ZWORK, ZMEAN, TZHALO2_VT%HALO2%WEST, TZHALO2_VT%HALO2%EAST, &
                           ZFPOS1(:,:,:), ZFPOS2(:,:,:), ZFNEG1(:,:,:), ZFNEG2(:,:,:),             &
                           ZBPOS1(:,:,:), ZBPOS2(:,:,:), ZBNEG1(:,:,:), ZBNEG2(:,:,:),             &
                           ZOMP1(:,:,:),  ZOMP2(:,:,:),  ZOMN1(:,:,:),  ZOMN2(:,:,:)               )
    CALL DXF_DEVICE(ZMEAN,ZWORK)
!$acc kernels
    PRVS(:,:,:) = PRVS(:,:,:) - ZWORK(:,:,:)
!$acc end kernels
!   
    CALL MYF_DEVICE(PRVCT,ZWORK)
    CALL ADVEC_WENO_K_2_VY(HLBCY, PVT, ZWORK, ZMEAN, TZHALO2_VT%HALO2%NORTH, TZHALO2_VT%HALO2%SOUTH, &
                           ZFPOS1(:,:,:), ZFPOS2(:,:,:), ZFNEG1(:,:,:), ZFNEG2(:,:,:),               &
                           ZBPOS1(:,:,:), ZBPOS2(:,:,:), ZBNEG1(:,:,:), ZBNEG2(:,:,:),               &
                           ZOMP1(:,:,:),  ZOMP2(:,:,:),  ZOMN1(:,:,:),  ZOMN2(:,:,:)                 )
    CALL DYM_DEVICE(ZMEAN,ZWORK)
!$acc kernels
    PRVS(:,:,:) = PRVS(:,:,:) - ZWORK(:,:,:)
!$acc end kernels
!
!    PRVS = PRVS - DZF(WENO_K_2_MZ(PVT, MYM(PRWCT)))
    CALL MYM_DEVICE(PRWCT,ZWORK)
    CALL WENO_K_2_MZ(PVT, ZWORK, ZMEAN,                                          &
                     ZFPOS1(:,:,:), ZFPOS2(:,:,:), ZFNEG1(:,:,:), ZFNEG2(:,:,:), &
                     ZBPOS1(:,:,:), ZBPOS2(:,:,:), ZBNEG1(:,:,:), ZBNEG2(:,:,:), &
                     ZOMP1(:,:,:),  ZOMP2(:,:,:),  ZOMN1(:,:,:),  ZOMN2(:,:,:)   )
    CALL DZF_DEVICE( ZMEAN, ZWORK )
!$acc kernels
    PRVS(:,:,:) = PRVS(:,:,:) - ZWORK(:,:,:)
!$acc end kernels
  END IF
!
! W component
!
!  ZWORK = MZM(PRUCT)
  CALL MZM_DEVICE(PRUCT,ZWORK)
  CALL ADVEC_WENO_K_2_MX(HLBCX, PWT, ZWORK, ZMEAN, TZHALO2_WT%HALO2%WEST, TZHALO2_WT%HALO2%EAST, &
                         ZFPOS1(:,:,:), ZFPOS2(:,:,:), ZFNEG1(:,:,:), ZFNEG2(:,:,:),             &
                         ZBPOS1(:,:,:), ZBPOS2(:,:,:), ZBNEG1(:,:,:), ZBNEG2(:,:,:),             &
                         ZOMP1(:,:,:),  ZOMP2(:,:,:),  ZOMN1(:,:,:),  ZOMN2(:,:,:)               )
  CALL DXF_DEVICE(ZMEAN,ZWORK)
!$acc kernels
  PRWS(:,:,:) = PRWS(:,:,:) - ZWORK(:,:,:)
!$acc end kernels
!
  IF (.NOT.L2D) THEN
!    ZWORK = MZM(PRVCT)
    CALL MZM_DEVICE(PRVCT,ZWORK)
    CALL ADVEC_WENO_K_2_MY(HLBCY, PWT, ZWORK, ZMEAN, TZHALO2_WT%HALO2%NORTH, TZHALO2_WT%HALO2%SOUTH, &
                           ZFPOS1(:,:,:), ZFPOS2(:,:,:), ZFNEG1(:,:,:), ZFNEG2(:,:,:),               &
                           ZBPOS1(:,:,:), ZBPOS2(:,:,:), ZBNEG1(:,:,:), ZBNEG2(:,:,:),               &
                           ZOMP1(:,:,:),  ZOMP2(:,:,:),  ZOMN1(:,:,:),  ZOMN2(:,:,:)                 )
    CALL DYF_DEVICE(ZMEAN,ZWORK)
!$acc kernels
    PRWS(:,:,:) = PRWS(:,:,:) - ZWORK(:,:,:)
!$acc end kernels
  END IF
!
!  PRWS = PRWS - DZM(WENO_K_2_WZ(PWT,MZF(PRWCT)))
  CALL MZF_DEVICE( PRWCT, ZWORK )
  CALL WENO_K_2_WZ(PWT, ZWORK, ZMEAN,                                          &
                   ZFPOS1(:,:,:), ZFPOS2(:,:,:), ZFNEG1(:,:,:), ZFNEG2(:,:,:), &
                   ZBPOS1(:,:,:), ZBPOS2(:,:,:), ZBNEG1(:,:,:), ZBNEG2(:,:,:), &
                   ZOMP1(:,:,:),  ZOMP2(:,:,:),  ZOMN1(:,:,:),  ZOMN2(:,:,:)   )
  CALL DZM_DEVICE( ZMEAN, ZWORK )
!$acc kernels
  PRWS(:,:,:) = PRWS(:,:,:) - ZWORK(:,:,:)
!$acc end kernels
!
  !Release all memory allocated with MNH_MEM_GET calls since last call to MNH_MEM_POSITION_PIN
  CALL MNH_MEM_RELEASE()
#endif
!
!
CASE(5) ! WENO 5
#ifndef MNH_OPENACC
!
! U component
!
  ZWORK = MXF(PRUCT)
  CALL ADVEC_WENO_K_3_UX(HLBCX, PUT, ZWORK, ZMEAN) 
  CALL GET_HALO(ZMEAN)! Update HALO 
  PRUS = PRUS - DXM(ZMEAN)
!
  IF (.NOT.L2D) THEN! 3D Case
   ZWORK = MXM(PRVCT)     
   CALL ADVEC_WENO_K_3_MY(HLBCY, PUT, ZWORK, ZMEAN)
   CALL GET_HALO(ZMEAN)! Update HALO 
   PRUS = PRUS - DYF(ZMEAN)
  END IF
!
  ZMEAN = WENO_K_3_MZ(PUT, MXM(PRWCT))
  CALL GET_HALO(ZMEAN)! Update HALO - maybe not necessary (T.Lunet)
  PRUS = PRUS - DZF(ZMEAN)
!
! V component, only called in 3D case
!
  IF (.NOT.L2D) THEN
!
    ZWORK = MYM(PRUCT)
    CALL ADVEC_WENO_K_3_MX(HLBCX, PVT, ZWORK, ZMEAN) 
    CALL GET_HALO(ZMEAN)! Update HALO 
    PRVS = PRVS - DXF(ZMEAN)
!
    ZWORK = MYF(PRVCT)
    CALL ADVEC_WENO_K_3_VY(HLBCY, PVT, ZWORK, ZMEAN)
    CALL GET_HALO(ZMEAN)! Update HALO 
    PRVS = PRVS - DYM(ZMEAN)
!
    ZMEAN = WENO_K_3_MZ(PVT, MYM(PRWCT))
    CALL GET_HALO(ZMEAN)! Update HALO - maybe not necessary (T.Lunet)
    PRVS = PRVS - DZF(ZMEAN)
!
  END IF
!
! W component
!
  ZWORK = MZM(PRUCT)
  CALL ADVEC_WENO_K_3_MX(HLBCX, PWT, ZWORK, ZMEAN)
  CALL GET_HALO(ZMEAN)! Update HALO
  PRWS = PRWS - DXF(ZMEAN)
!
  IF (.NOT.L2D) THEN! 3D Case
    ZWORK = MZM(PRVCT)
    CALL ADVEC_WENO_K_3_MY(HLBCY, PWT, ZWORK, ZMEAN)
    CALL GET_HALO(ZMEAN)! Update HALO
    PRWS = PRWS - DYF(ZMEAN)
  END IF
!
  ZMEAN = WENO_K_3_WZ(PWT,MZF(PRWCT))
  CALL GET_HALO(ZMEAN)! Update HALO - maybe not necessary (T.Lunet)
  PRWS = PRWS - DZM(ZMEAN)
#else
  !Pin positions in the pools of MNH memory
  CALL MNH_MEM_POSITION_PIN()

  CALL MNH_MEM_GET( ZFPOS1, IIU, IJU, IKU )
  CALL MNH_MEM_GET( ZFPOS2, IIU, IJU, IKU )
  CALL MNH_MEM_GET( ZFPOS3, IIU, IJU, IKU )
  CALL MNH_MEM_GET( ZFNEG1, IIU, IJU, IKU )
  CALL MNH_MEM_GET( ZFNEG2, IIU, IJU, IKU )
  CALL MNH_MEM_GET( ZFNEG3, IIU, IJU, IKU )
  CALL MNH_MEM_GET( ZBPOS1, IIU, IJU, IKU )
  CALL MNH_MEM_GET( ZBPOS2, IIU, IJU, IKU )
  CALL MNH_MEM_GET( ZBPOS3, IIU, IJU, IKU )
  CALL MNH_MEM_GET( ZBNEG1, IIU, IJU, IKU )
  CALL MNH_MEM_GET( ZBNEG2, IIU, IJU, IKU )
  CALL MNH_MEM_GET( ZBNEG3, IIU, IJU, IKU )
  CALL MNH_MEM_GET( ZOMP1,  IIU, IJU, IKU )
  CALL MNH_MEM_GET( ZOMP2,  IIU, IJU, IKU )
  CALL MNH_MEM_GET( ZOMP3,  IIU, IJU, IKU )
  CALL MNH_MEM_GET( ZOMN1,  IIU, IJU, IKU )
  CALL MNH_MEM_GET( ZOMN2,  IIU, IJU, IKU )
  CALL MNH_MEM_GET( ZOMN3,  IIU, IJU, IKU )
!
! U component
!
  CALL MXF_DEVICE(PRUCT,ZWORK)
  CALL ADVEC_WENO_K_3_UX(HLBCX, PUT, ZWORK, ZMEAN,                    &
                         ZFPOS1(:,:,:), ZFPOS2(:,:,:), ZFPOS3(:,:,:), &
                         ZFNEG1(:,:,:), ZFNEG2(:,:,:), ZFNEG3(:,:,:), &
                         ZBPOS1(:,:,:), ZBPOS2(:,:,:), ZBPOS3(:,:,:), &
                         ZBNEG1(:,:,:), ZBNEG2(:,:,:), ZBNEG3(:,:,:), &
                         ZOMP1(:,:,:),  ZOMP2(:,:,:),  ZOMP3(:,:,:),  &
                         ZOMN1(:,:,:),  ZOMN2(:,:,:),  ZOMN3(:,:,:)   )
  CALL GET_HALO_D(ZMEAN)! Update HALO 
  CALL DXM_DEVICE(ZMEAN,ZWORK)
!$acc kernels
  PRUS(:,:,:) = PRUS(:,:,:) - ZWORK(:,:,:)
!$acc end kernels
!
  IF (.NOT.L2D) THEN
    CALL MXM_DEVICE(PRVCT,ZWORK)
    CALL ADVEC_WENO_K_3_MY(HLBCY, PUT, ZWORK, ZMEAN,                    &
                           ZFPOS1(:,:,:), ZFPOS2(:,:,:), ZFPOS3(:,:,:), &
                           ZFNEG1(:,:,:), ZFNEG2(:,:,:), ZFNEG3(:,:,:), &
                           ZBPOS1(:,:,:), ZBPOS2(:,:,:), ZBPOS3(:,:,:), &
                           ZBNEG1(:,:,:), ZBNEG2(:,:,:), ZBNEG3(:,:,:), &
                           ZOMP1(:,:,:),  ZOMP2(:,:,:),  ZOMP3(:,:,:),  &
                           ZOMN1(:,:,:),  ZOMN2(:,:,:),  ZOMN3(:,:,:)   )
    CALL GET_HALO_D(ZMEAN)! Update HALO 
    CALL DYF_DEVICE(ZMEAN,ZWORK)
!$acc kernels
    PRUS(:,:,:) = PRUS(:,:,:) - ZWORK(:,:,:)
!$acc end kernels
  END IF
!
  CALL MXM_DEVICE(PRWCT,ZWORK)
  CALL WENO_K_3_MZ(PUT,ZWORK,ZMEAN,                             &
                   ZFPOS1(:,:,:), ZFPOS2(:,:,:), ZFPOS3(:,:,:), &
                   ZFNEG1(:,:,:), ZFNEG2(:,:,:), ZFNEG3(:,:,:), &
                   ZBPOS1(:,:,:), ZBPOS2(:,:,:), ZBPOS3(:,:,:), &
                   ZBNEG1(:,:,:), ZBNEG2(:,:,:), ZBNEG3(:,:,:), &
                   ZOMP1(:,:,:),  ZOMP2(:,:,:),  ZOMP3(:,:,:),  &
                   ZOMN1(:,:,:),  ZOMN2(:,:,:),  ZOMN3(:,:,:)   )
  CALL GET_HALO_D(ZMEAN)! Update HALO - maybe not necessary (T.Lunet)
  CALL DZF_DEVICE( ZMEAN, ZWORK )
!$acc kernels
  PRUS(:,:,:) = PRUS(:,:,:) - ZWORK(:,:,:)
!$acc end kernels
!
! V component
!
  IF (.NOT.L2D) THEN
    CALL MYM_DEVICE(PRUCT,ZWORK)
    CALL ADVEC_WENO_K_3_MX(HLBCX, PVT, ZWORK, ZMEAN,                    &
                           ZFPOS1(:,:,:), ZFPOS2(:,:,:), ZFPOS3(:,:,:), &
                           ZFNEG1(:,:,:), ZFNEG2(:,:,:), ZFNEG3(:,:,:), &
                           ZBPOS1(:,:,:), ZBPOS2(:,:,:), ZBPOS3(:,:,:), &
                           ZBNEG1(:,:,:), ZBNEG2(:,:,:), ZBNEG3(:,:,:), &
                           ZOMP1(:,:,:),  ZOMP2(:,:,:),  ZOMP3(:,:,:),  &
                           ZOMN1(:,:,:),  ZOMN2(:,:,:),  ZOMN3(:,:,:)   )
    CALL GET_HALO_D(ZMEAN)! Update HALO 
    CALL DXF_DEVICE(ZMEAN,ZWORK)
!$acc kernels
    PRVS(:,:,:) = PRVS(:,:,:) - ZWORK(:,:,:)
!$acc end kernels
!   
    CALL MYF_DEVICE(PRVCT,ZWORK)
    CALL ADVEC_WENO_K_3_VY(HLBCY, PVT, ZWORK, ZMEAN,                    &
                           ZFPOS1(:,:,:), ZFPOS2(:,:,:), ZFPOS3(:,:,:), &
                           ZFNEG1(:,:,:), ZFNEG2(:,:,:), ZFNEG3(:,:,:), &
                           ZBPOS1(:,:,:), ZBPOS2(:,:,:), ZBPOS3(:,:,:), &
                           ZBNEG1(:,:,:), ZBNEG2(:,:,:), ZBNEG3(:,:,:), &
                           ZOMP1(:,:,:),  ZOMP2(:,:,:),  ZOMP3(:,:,:),  &
                           ZOMN1(:,:,:),  ZOMN2(:,:,:),  ZOMN3(:,:,:)   )
    CALL GET_HALO_D(ZMEAN)! Update HALO 
    CALL DYM_DEVICE(ZMEAN,ZWORK)
!$acc kernels
    PRVS(:,:,:) = PRVS(:,:,:) - ZWORK(:,:,:)
!$acc end kernels
!
    CALL MYM_DEVICE(PRWCT,ZWORK)
    CALL WENO_K_3_MZ(PVT, ZWORK, ZMEAN,                           &
                     ZFPOS1(:,:,:), ZFPOS2(:,:,:), ZFPOS3(:,:,:), &
                     ZFNEG1(:,:,:), ZFNEG2(:,:,:), ZFNEG3(:,:,:), &
                     ZBPOS1(:,:,:), ZBPOS2(:,:,:), ZBPOS3(:,:,:), &
                     ZBNEG1(:,:,:), ZBNEG2(:,:,:), ZBNEG3(:,:,:), &
                     ZOMP1(:,:,:),  ZOMP2(:,:,:),  ZOMP3(:,:,:),  &
                     ZOMN1(:,:,:),  ZOMN2(:,:,:),  ZOMN3(:,:,:)   )
    CALL GET_HALO_D(ZMEAN)! Update HALO - maybe not necessary (T.Lunet)
    CALL DZF_DEVICE( ZMEAN, ZWORK )
!$acc kernels
    PRVS(:,:,:) = PRVS(:,:,:) - ZWORK(:,:,:)
!$acc end kernels
  END IF
!
! W component
!
  CALL MZM_DEVICE(PRUCT,ZWORK)
  CALL ADVEC_WENO_K_3_MX(HLBCX, PWT, ZWORK, ZMEAN,                    &
                         ZFPOS1(:,:,:), ZFPOS2(:,:,:), ZFPOS3(:,:,:), &
                         ZFNEG1(:,:,:), ZFNEG2(:,:,:), ZFNEG3(:,:,:), &
                         ZBPOS1(:,:,:), ZBPOS2(:,:,:), ZBPOS3(:,:,:), &
                         ZBNEG1(:,:,:), ZBNEG2(:,:,:), ZBNEG3(:,:,:), &
                         ZOMP1(:,:,:),  ZOMP2(:,:,:),  ZOMP3(:,:,:),  &
                         ZOMN1(:,:,:),  ZOMN2(:,:,:),  ZOMN3(:,:,:)   )
  CALL GET_HALO_D(ZMEAN)! Update HALO 
  CALL DXF_DEVICE(ZMEAN,ZWORK)
!$acc kernels
  PRWS(:,:,:) = PRWS(:,:,:) - ZWORK(:,:,:)
!$acc end kernels
!
  IF (.NOT.L2D) THEN
  CALL MZM_DEVICE(PRVCT,ZWORK)
    CALL ADVEC_WENO_K_3_MY(HLBCY, PWT, ZWORK, ZMEAN,                    &
                           ZFPOS1(:,:,:), ZFPOS2(:,:,:), ZFPOS3(:,:,:), &
                           ZFNEG1(:,:,:), ZFNEG2(:,:,:), ZFNEG3(:,:,:), &
                           ZBPOS1(:,:,:), ZBPOS2(:,:,:), ZBPOS3(:,:,:), &
                           ZBNEG1(:,:,:), ZBNEG2(:,:,:), ZBNEG3(:,:,:), &
                           ZOMP1(:,:,:),  ZOMP2(:,:,:),  ZOMP3(:,:,:),  &
                           ZOMN1(:,:,:),  ZOMN2(:,:,:),  ZOMN3(:,:,:)   )
    CALL GET_HALO_D(ZMEAN)! Update HALO 
    CALL DYF_DEVICE(ZMEAN,ZWORK)
!$acc kernels
    PRWS(:,:,:) = PRWS(:,:,:) - ZWORK(:,:,:)
!$acc end kernels
  END IF
!
  CALL MZF_DEVICE( PRWCT, ZWORK )
  CALL WENO_K_3_WZ(PWT,ZWORK,ZMEAN,                             &
                   ZFPOS1(:,:,:), ZFPOS2(:,:,:), ZFPOS3(:,:,:), &
                   ZFNEG1(:,:,:), ZFNEG2(:,:,:), ZFNEG3(:,:,:), &
                   ZBPOS1(:,:,:), ZBPOS2(:,:,:), ZBPOS3(:,:,:), &
                   ZBNEG1(:,:,:), ZBNEG2(:,:,:), ZBNEG3(:,:,:), &
                   ZOMP1(:,:,:),  ZOMP2(:,:,:),  ZOMP3(:,:,:),  &
                   ZOMN1(:,:,:),  ZOMN2(:,:,:),  ZOMN3(:,:,:)   )
  CALL GET_HALO_D(ZMEAN)! Update HALO - maybe not necessary (T.Lunet)
  CALL DZM_DEVICE( ZMEAN, ZWORK )
!$acc kernels
  PRWS(:,:,:) = PRWS(:,:,:) - ZWORK(:,:,:)
!$acc end kernels
!
  !Release all memory allocated with MNH_MEM_GET calls since last call to MNH_MEM_POSITION_PIN
  CALL MNH_MEM_RELEASE()
#endif
!
!
END SELECT
!             ---------------------------------
! acc update self(PRUS,PRVS,PRWS)

!$acc end data

#ifdef MNH_OPENACC
!Release all memory allocated with MNH_MEM_GET calls since beginning of subroutine
CALL MNH_MEM_RELEASE()
#endif

IF (MPPDB_INITIALIZED) THEN
  CALL MPPDB_CHECK(PRUS,"ADVECUVW_WENO_K end:PRUS")
  CALL MPPDB_CHECK(PRVS,"ADVECUVW_WENO_K end:PRVS")
  CALL MPPDB_CHECK(PRWS,"ADVECUVW_WENO_K end:PRWS")
END IF


END SUBROUTINE ADVECUVW_WENO_K
