!=== COPYRIGHT AND LICENSE STATEMENT ===
!
!  This file is part of the TensorProductMultigrid code.
!  
!  (c) The copyright relating to this work is owned jointly by the
!  Crown, Met Office and NERC [2014]. However, it has been created
!  with the help of the GungHo Consortium, whose members are identified
!  at https://puma.nerc.ac.uk/trac/GungHo/wiki .
!  
!  Main Developer: Eike Mueller
!  
!  TensorProductMultigrid is free software: you can redistribute it and/or
!  modify it under the terms of the GNU Lesser General Public License as
!  published by the Free Software Foundation, either version 3 of the
!  License, or (at your option) any later version.
!  
!  TensorProductMultigrid is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU Lesser General Public License for more details.
!  
!  You should have received a copy of the GNU Lesser General Public License
!  along with TensorProductMultigrid (see files COPYING and COPYING.LESSER).
!  If not, see <http://www.gnu.org/licenses/>.
!
!=== COPYRIGHT AND LICENSE STATEMENT ===


!==================================================================
!
!  Main program for multigrid solver code for Helmholtz/Poisson
!  equation, discretised in the cell centred finite volume scheme
!
!    Eike Mueller, University of Bath, Feb 2012
!
!==================================================================

!==================================================================
! Main program
!==================================================================

program mg_main

  use discretisation
  use parameters
  use datatypes
  use multigrid
  use conjugategradient
  use solver
  use profiles
  use messages
  use communication
  use timer
#ifndef MNH
  use mpi
#else
  use modd_mpif
#endif

  use mode_mg_read_param
  use mode_mg

  implicit none

  call mg_init()

  ! Initialise ghosts in initial solution, as mg_solve assumes that they
  ! are up-to-date
  call haloswap(mg_param%n_lev,pproc,xu_fine)

  ! Solve using multigrid
  call initialise_timer(t_solve,"t_solve")
  call start_timer(t_solve)
  comm_measuretime = .True.
#ifdef MEASUREHALOSWAP
  call measurehaloswap()
#else
  call mg_solve(xb_fine,xu_fine,solver_param)
#endif
  comm_measuretime = .False.
  call finish_timer(t_solve)

call mg_finalize()

end program mg_main
